/*
 * The contents of this file are subject to the JOnAS Public License Version 
 * 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License on the JOnAS web site.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.
 * See the License for the specific terms governing rights and limitations under 
 * the License.
 *
 * The Original Code is JOnAS application server code released July 1999.
 *
 * The Initial Developer of the Original Code is Bull S.A.
 * The Original Code and portions created by Bull S.A. are
 *    Copyright (C) 1999 Bull S.A. All Rights Reserved.
 *
 * Contributor(s): ______________________________________.
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */


package org.objectweb.jonas.container; 

import java.lang.IllegalStateException;
import java.util.Properties;
import javax.ejb.EntityBean;
import javax.ejb.EJBContext;
import javax.ejb.EJBHome;
import javax.ejb.EJBObject;
import javax.ejb.EntityContext;
import javax.transaction.Transaction;
import javax.transaction.UserTransaction;
import org.objectweb.jonas.common.Trace;


/**
 * This class implements javax.ejb.EntityContext interface.
 *
 * This class implements the  context associated to an EJBObject Entity
 * in order to allow concurrent access from multiple transactions 
 */

public class EntityContextImpl extends EJBContextImpl implements EntityContext {

    static private final int trace = Trace.DB_12;

    private JBeanEntity ejbObject  = null; // the remote object
    private EntityBean enterpriseBean = null; // the bean instance

    EntitySynchroImpl synchro = null; // Synchro associated with the bean instance
    Transaction clientTransaction = null; // Tx to be resumed at postinvoke
    boolean mustBeCommitted = false; // must commit current Tx at postinvoke

    // ------------------------------------------------------------------
    // constructors
    // ------------------------------------------------------------------

    /**
     * Constructs an EntityContext 
     * 
     * @param obj  the EJB Object created
     * @param home the enterprise bean's home interface
     * @param eb   the Enterprise Bean
     * @param sync the Synchronization Object
     *
     */
    public EntityContextImpl(JBeanEntity obj, 
			     JBeanHome home, 
			     EntityBean eb, 
			     EntitySynchroImpl sync) {
	super(home);
	Trace.outln(trace, "EntityContextImpl.EntityContextImpl(EJBObject "+obj+" , EJBHome " +home+")");
	ejbObject = obj;
	enterpriseBean = eb;
	synchro = sync;
    }
  

    /**
     * Constructs a minimal EntityContext used by Finder Methods
     * in interposition classes
     */
    public EntityContextImpl(JBeanHome home) {
	super(home);
	Trace.outln(trace, "EntityContextImpl.EntityContextImpl(EJBHome " +home+")");
    }

    // ------------------------------------------------------------------
    // EntityContext implementation
    // ------------------------------------------------------------------


    /**
     * Obtains a reference to the EJB object that is currently associated with the instance. 
     *  An instance of a session enterprise Bean can call this method at anytime between the
     *   ejbCreate() and ejbRemove() methods, including from within the ejbCreate() and
     *   ejbRemove() methods.
     *  An instance can use this method, for example, when it wants to pass a reference to 
     *   itself in a method argument or result.
     * @return The EJB object currently associated with the instance. 
     * @exception IllegalStateException: Thrown if the instance invokes this 
     *            method while the instance is in a state that does not allow the 
     *            instance to invoke this method.
     */
    public  EJBObject getEJBObject() throws IllegalStateException {
	Trace.outln(trace, "EntityContextImpl.getEJBObject()");

	if (synchro.isMarkedRemoved())
	    throw new IllegalStateException("EJB Object is removed");
	return ejbObject;
    }


    /**
     * Obtains the primary key of the EJB object that is currently associated 
     *  with this instance. 
     * An instance of an entity enterprise Bean can call this method only when the 
     *  instance is associated with an EJB object identity, i.e. in the ejbActivate, 
     *  ejbPassivate, ejbPostCreate method, ejbRemove, ejbLoad, ejbStore, and 
     *  business methods. 
     * Note: The result of this method is the same as the result of 
     *  getEJBObject().getPrimaryKey(). 
     *
     * @return  The EJB object currently associated with the instance. 
     *
     * @exception  IllegalStateException  Thrown if the instance invokes this 
     *             method while the instance is in a state that does not allow 
     *             the instance to invoke this method.     
     */
    public Object getPrimaryKey() throws IllegalStateException {
	Trace.outln(trace, "EntityContextImpl.getPrimaryKey() ");
	try{
	    Object pk = ejbObject.getPrimaryKey();
	    if (pk == null)
		throw new IllegalStateException();
	    return pk;
	} catch(java.rmi.RemoteException e) {
	    Trace.errln("EntityContextImpl.getPrimaryKey() RemoteException caught :" +e);
	    throw new IllegalStateException();
	}
    }


    // ------------------------------------------------------------------
    // Other methods
    // ------------------------------------------------------------------

    /**
     * Sets the primarykey of the Bean.
     * the primary key is known after ejbCreate method on the  bean
     * and kept in the EJBObject
     *
     * @param pk - the primaryKey object
     *
     */
    public void setPrimaryKey(Object pk){
	Trace.outln(trace, "EntityContextImpl.setPrimaryKey("+pk+")");
	ejbObject.setPrimaryKey(pk);
    }


   /**
     * Returns the bean instance of this context
     *
     * @return the bean instance
     *
     */
    public EntityBean getInstance() {
	return enterpriseBean;
    }
}
