/*
 * The contents of this file are subject to the JOnAS Public License Version 
 * 1.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License on the JOnAS web site.
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied.
 * See the License for the specific terms governing rights and limitations under 
 * the License.
 *
 * The Original Code is JOnAS application server code released July 1999.
 *
 * The Initial Developer of the Original Code is Bull S.A.
 * The Original Code and portions created by Bull S.A. are
 *    Copyright (C) 1999 Bull S.A. All Rights Reserved.
 *
 * Contributor(s): ______________________________________.
 *
 * --------------------------------------------------------------------------
 * $Id$
 * --------------------------------------------------------------------------
 */


package org.objectweb.jonas.tools;

import java.lang.reflect.Method;
import javax.ejb.deployment.SessionDescriptor;
import org.objectweb.jonas.common.Trace;
import org.objectweb.jonas.lib.DD;
import org.objectweb.jonas.lib.InvalidBeanException;
import org.objectweb.jonas.lib.JavaType;


/**
 *  This class allows to check the validity of a Session Enterprise Java Bean.
 *  It extends the CheckBean class.
 */

public class CheckSession extends CheckBean {

    static final int trace  = Trace.DB_19;

    int stateManagementType;

    /**
     * Construct a CheckSession from a deployment descriptor of a Session Enterprise Java Bean.
     * @param ddesc deployment descriptor of the session bean
     * @exception InvalidBeanException in fatal error case
     */
    public CheckSession(DD ddesc) throws InvalidBeanException {
	super(ddesc);
	stateManagementType = ejbDD.getSessionDescriptor().getStateManagementType();
    }

    /**
     * Returns true if the Session Bean is valid, false otherwise
     * @exception InvalidBeanException in fatal error case
     */
    public boolean check() throws InvalidBeanException {
	boolean isValid = true;
	if (!super.check()) {
	    isValid = false;
	}
	if (!checkIRemote()) {
	    isValid = false;
	}
	if (!checkIHome()) {
	    isValid = false;
	}
	if (!checkBean()) {
	    isValid = false;
	}
	return(isValid);
    }

    /**
     * Returns true if the Interface Remote class is valid, false otherwise.
     */
    private boolean checkIRemote() {
	boolean isValid = true;
	return(isValid);
    }

    /**
     * Returns true if the Interface Home class is valid, false otherwise.
     */
    private boolean checkIHome() {
	boolean isValid = true;

	Method [] methods = ejbHomeClass.getMethods();
	int nbCreate = 0;
	for (int i=0; i<methods.length; i++) {
	    // Skip the methods defined in the 'javax.ejb.EJBHome' interface
	    Method meth = methods[i];
	    if (!meth.getDeclaringClass().getName().equals("javax.ejb.EJBHome")) {
		if (meth.getName().equals("create")) {
		    nbCreate++;
		    if (stateManagementType == SessionDescriptor.STATELESS_SESSION) {
			Class [] params = meth.getParameterTypes();
			if (params.length != 0) {
			    Trace.errln("Error: The home interface of a stateless session Bean must have");
			    Trace.errln("       a 'create' method that takes no arguments");
			    isValid = false;	
			}
		    }
		} else {
		    Trace.errln("Error: Only 'create' methods can be defined in a Session Bean's");
		    Trace.errln("       Home Interface");
		    Trace.errln("  Method: "+meth.toString());
		    isValid = false;
		}
	    }
	}

	if (nbCreate < 1) {
	    Trace.errln("Error: A Session Bean's Home Interface defines at least one 'create' method");
	    isValid = false;
	}

	if (stateManagementType == SessionDescriptor.STATELESS_SESSION) {
	    if (nbCreate != 1) {
		Trace.errln("Error: The Home Interface of a stateless session Bean must have one");
		Trace.errln("       and only one 'create' method");
		isValid = false;
	    }
	}

	return(isValid);
    }

    /**
     * Returns true if the Session Bean class is valid, false otherwise.
     */
    private boolean checkBean() {
	boolean isValid = true;

	if (!JavaType.implementsInterface(ejbBeanClass, javax.ejb.SessionBean.class)) {
	    Trace.errln("Error: The Enterprise Session Bean class must implement the interface");
	    Trace.errln("       'javax.ejb.SessionBean'");
	    isValid = false;
	}

	Method [] methods = ejbBeanClass.getMethods();
	for (int i=0; i<methods.length; i++) {
	    Method meth = methods[i];
	    if (meth.getName().equals("ejbCreate")) {
		if (!JavaType.isVoid(meth.getReturnType())) {
		    Trace.errln("Error: The return type of an 'ejbCreate' method of a Session");
		    Trace.errln("       Bean must be 'void'");
		    Trace.errln("  Method: "+meth.toString());
		    isValid = false;
		}
	    }
	}

	if (stateManagementType == SessionDescriptor.STATELESS_SESSION) {
	    if (JavaType.implementsInterface(ejbBeanClass,
					     javax.ejb.SessionSynchronization.class)) {
		Trace.errln("Error: A stateless Session must not implement the interface");
		Trace.errln("       'javax.ejb.SessionSynchronization'");
		isValid = false;
	    }
	}

	if (ejbDD.getDeploymentDescriptor().isReentrant()) {
	    Trace.errln("Error: Only Entity Beans can be defined as re-entrant");
	    isValid = false;
	}
	return(isValid);
    }

}

